//================================================================
//MARK: Class Body
//================================================================

class Body {

  //-------------------------------------------------
  //Public Properties
  //-------------------------------------------------
  
  float segmentCount;
  ArrayList<Segment> segments;

  float totalLength = 0.0;
  float startPos, endPos;
  int waitTimer;

  float _cameraEasedY, _cameraEasedScale;
  PVector bodyStartPoint, bodyEndPoint; //should be readonly

  //-------------------------------------------------
  //Private Properties
  //-------------------------------------------------

  private Ramp startPosRamp, endPosRamp;
  private boolean startEndFlip = false;
  private Rune restRune;
  private PVector restPoint; 
  private boolean resting;
  private boolean waitingForRuneToHide;
  private int firstSegmentIndex, lastSegmentIndex;
  private color bodyColor;
  private float totalBodyLength;

  private PVector startRampValueDuration, endRampValueDuration;

  Body(int sCount, color _color) {
    segmentCount = sCount;

    endPosRamp = new Ramp();
    endPosRamp.sCurvePower = 6;

    startPosRamp = new Ramp();
    startPosRamp.sCurvePower = 6;

    bodyColor = _color;

    segments = new ArrayList<Segment>();

    int[] tendrilCount = new int[sCount];
    int[] rt = new int[sCount];
    for (int i=0; i<rt.length; i++) {
      rt[i] = i;
    }
    
    float randomPhase = random(TWO_PI);
    float freq = random(1.5, 2.5);
    for (int i=0; i<tendrilCount.length; i++) {
      float norm = float(i) / tendrilCount.length;
      norm = abs(sin(norm * TWO_PI * freq + randomPhase));
      tendrilCount[i] = int(norm * norm * random(5, 20));
    }

    //growing plant top (will flip during draw) random
    PVector origin = new PVector(0, 0);
    PVector lastPoint = origin;
    bodyStartAnchorPoint = origin;
    for (int i=0; i<segmentCount; i++) {
      float x = lastPoint.x + (random(-kMaxRandomWalkWidth, kMaxRandomWalkWidth));
      x = max(min(x, width / 2.0), -width / 2.0);
      float y = lastPoint.y + 50;
      float l = random(kSegmentMaximumLength + kSegmentMinimumLength) + kSegmentMinimumLength;
      PVector rand = new PVector(x, y);
      Segment seg = new Segment(l, lastPoint, rand, tendrilCount[i]);
      segments.add(seg);
      totalLength += segments.get(i).len;
      lastPoint = segments.get(i).pointInSegment(1.0);

      seg.setColor(bodyColor);
      seg.setColorBrightness(brightness(bodyColor) * 0.35);

      //update previous cp1
      if (i > 0) {
        Segment s = segments.get(i);
        Segment prevS = segments.get(i-1);

        //get this segment's cp0, rotate it by theta, and set the previous to its opposite
        float cp0Len = sqrt(s.cp0.x*s.cp0.x + s.cp0.y*s.cp0.y);
        float theta = atan2(s.cp0.y, s.cp0.x) + PI;
        float dTheta = s.theta - prevS.theta;
        theta += dTheta;
        float newX = (cos(theta) * cp0Len) + prevS.len;
        float newY = (sin(theta) * cp0Len);
        prevS.cp1 = new PVector(newX, newY);
      }
    }
    totalBodyLength = totalLength;

    Segment last = segments.get(segments.size() - 1);
    float x = lastPoint.x + cos(last.theta) * last.len;
    float y = lastPoint.y + sin(last.theta) * last.len;
    bodyEndAnchorPoint = new PVector(x, y);

    float lengthSum = 0.0;
    for (int i=0; i<segments.size(); i++) {
      Segment s = segments.get(i);
      s.bodyPosition = lengthSum;
      lengthSum += s.len;
      s.totalBodyLength = totalLength;
    }

    startPos = 0.0;
    endPos = 0.0;

    //we use {-1, -1} to indicate a 'null'
    startRampValueDuration = new PVector(-1, -1);
    endRampValueDuration = new PVector(-1, -1);

    //set up rest rune
    color runeColor = color(hue(bodyColor), saturation(bodyColor), max(brightness(bodyColor), 150));
    restRune = new Rune(200, runeColor);

    Segment firstSegment = segments.get(firstSegmentIndex);
    Segment lastSegment = segments.get(lastSegmentIndex);
    bodyStartPoint = firstSegment.startPoint();
    bodyEndPoint = lastSegment.endPoint();

    restPoint = new PVector(
      (bodyStartPoint.x + bodyEndPoint.x) / 2.0, 
      (bodyStartPoint.y + bodyEndPoint.y) / 2.0
      );
    setResting(true);
  }

  void update() {
    if (kActive) {
      if (int(random(560)) == 0) {
        setResting(true);
      } else if (int(random(360)) == 0) {
        setResting(false);
      }
    }
    if (!resting) {
      if (millis() - waitTimer < 700) {
        waitingForRuneToHide = true;
        return;
      }

      if (waitingForRuneToHide) {
        startPosRamp.setValue(startPos);
        endPosRamp.setValue(endPos);
        waitingForRuneToHide = false;
      }

      boolean startAlmostFinished = startPosRamp.percentageComplete() > 0.8;
      boolean endAlmostFinished = endPosRamp.percentageComplete() > 0.8;
      if (startAlmostFinished &&
        endAlmostFinished) {
        if (startEndFlip) {
          startPosRamp.randomizeMode();
          startRampValueDuration.set(random(endPos), (random(500) + 2000) * kSpeedMultiplier);
        } else {
          endPosRamp.randomizeMode();
          float minLength = 100;
          float tep = min(totalLength, random(totalLength - startPos) + startPos + minLength);
          endRampValueDuration.set(tep, (random(500) + 2000) * kSpeedMultiplier);
        }
        startEndFlip = !startEndFlip;
      }
    }

    //check for ramps to start
    if (startRampValueDuration.y != -1) {
      startPosRamp.rampToValueOverDuration(startRampValueDuration.x, startRampValueDuration.y);
      startRampValueDuration.set(-1, -1);
    }
    if (endRampValueDuration.y != -1) {
      endPosRamp.rampToValueOverDuration(endRampValueDuration.x, endRampValueDuration.y);
      endRampValueDuration.set(-1, -1);
    }

    startPos = startPosRamp.update();
    endPos = endPosRamp.update();

    firstSegmentIndex = MAX_INT;
    lastSegmentIndex = -1;
    for (int i=0; i<segments.size(); i++) {
      Segment s = segments.get(i);
      boolean updated = s.update(startPos, endPos);
      if (updated) {
        if (i < firstSegmentIndex) {
          firstSegmentIndex = i;
        }
        if (i > lastSegmentIndex) {
          lastSegmentIndex = i;
        }
      }
    }

    if (lastSegmentIndex == -1 ||
      firstSegmentIndex == MAX_INT ||
      startPos > endPos) {
      //we're shrunk to zero, so...
      return;
    }

    Segment firstSegment = segments.get(firstSegmentIndex);
    Segment lastSegment = segments.get(lastSegmentIndex);
    bodyStartPoint = firstSegment.startPoint();
    bodyEndPoint = lastSegment.endPoint();
  }

  void drawBody(PGraphics pg) {
    if (resting && startPos == endPos) {
      restRune.showAtPosition(restPoint);
    }
    if (restRune.isShowing()) {
      restRune.updateDraw(pg);
    } else {
      for (int i=0; i<segments.size(); i++) {
        Segment s = segments.get(i);
        s.drawSegment(pg);
      }
    }
  }

  void setResting(boolean r) {
    if (r == resting) {
      return;
    }
    resting = r;
    if (!resting) {
      restRune.hide();
      waitTimer = millis();
    }
    if (resting) {
      float restPhase = (startPos + endPos) / 2.0;
      startPosRamp.rampToValueOverDuration(restPhase, kRestRampTime);
      endPosRamp.rampToValueOverDuration(restPhase, kRestRampTime);

      PVector point = pointAtPhase(restPhase / totalBodyLength);
      if (point != null) {
        restPoint = pointAtPhase(restPhase / totalBodyLength);
        startEndFlip = false;
      } else {
        println("null restPoint restPhase "+restPhase+" tbl "+totalBodyLength);
      }
    }
  }

  //-------------------------------------------------
  //Functions
  //-------------------------------------------------

  PVector pointAtPhase(float phase) {
    if (phase < 0.0 || phase > 1.0) {
      return null;
    }

    float phaseLength = totalBodyLength * phase;

    float runningBodyLength = 0.0;
    for (int i=0; i<segments.size(); i++) {
      Segment seg = segments.get(i);
      float startSegmentLength = runningBodyLength;
      runningBodyLength += seg.len;
      if (runningBodyLength > phaseLength) {
        //okay, this is the segment which has the phase
        float endSegmentLength = runningBodyLength;
        float segPhase = (phaseLength - startSegmentLength) / (endSegmentLength - startSegmentLength);
        PVector point = seg.pointInSegment(segPhase);
        return point;
      }
    }
    return null;
  }
  
  void scaleSegmentColorBrightness(float b) {
      for (int i=0; i<segments.size(); i++) {
        Segment seg = segments.get(i);
        seg.scaleColorBrightness(b);
      } 
  }
}