//================================================================
//MARK: Ramp
//================================================================

enum RampMode {
  RAMP_MODE_LINEAR, 
    RAMP_MODE_EASE_IN, 
    RAMP_MODE_EASE_OUT, 
    RAMP_MODE_S_CURVE
}

class Ramp {

  //================================================================
  //MARK: Public Properties
  //================================================================

  float[] valueArray;
  float value = 0.0;
  RampMode mode = RampMode.RAMP_MODE_LINEAR;
  float sCurvePower = 1;

  //================================================================
  //MARK: Private Properties
  //================================================================

  private float start;
  private float end;

  private float[] startArray;
  private float[] endArray;

  private float startTime;
  private float duration;
  private boolean running;


  //================================================================
  //MARK: Readonly Properties
  //================================================================

  float percentageComplete() {
    if (duration == 0) {
      return 1.0;
    }
    float m = millis();
    float p = (m - startTime) / duration;
    //println("percCom m st d "+p+" "+m+" "+startTime+" "+duration);
    return p;
  }

  //================================================================
  //MARK: Initialization
  //================================================================

  Ramp() {
    start = 0.0;
    end = 0.0;
    duration = 0;
    value = 0.0;
    startTime = 0;

    running = false;
  }

  //================================================================
  //MARK: Value Changing Functions
  //================================================================

  void rampToValueOverDuration(float v, float d) {
    if (d == 0) {
      setValue(v);
    } else {
      start = value;
      end = v;
      duration = d;
      start();
    }
  }

  void setValue(float v) {
    start = v;
    end = v;
    value = v;
    duration = 0;
    start();
  }

  void setValues(float[] v) {
    startArray = new float[v.length];
    for (int i=0; i<v.length; i++) {
      startArray[i] = v[i];
    }
    endArray = new float[v.length];
    for (int i=0; i<v.length; i++) {
      endArray[i] = v[i];
    }
    valueArray = new float[v.length];
    for (int i=0; i<v.length; i++) {
      valueArray[i] = v[i];
    }
    duration = 0;
    start();
  }

  void rampToValuesOverDuration(float[] v, float d) {
    if (d == 0) {
      setValues(v);
    } else {
      startArray = new float[v.length];
      for (int i=0; i<v.length; i++) {
        if (valueArray != null &&
          valueArray.length > i) {
          startArray[i] = valueArray[i];
        } else {
          startArray[i] = 0;
        }
      }
      endArray = new float[v.length];
      for (int i=0; i<v.length; i++) {
        endArray[i] = v[i];
      }
      valueArray = new float[v.length];
      for (int i=0; i<v.length; i++) {
        valueArray[i] = startArray[i];
      }
      duration = d;
      start();
    }
  }

  //================================================================
  //MARK: Start / Stop / Update
  //================================================================

  void start() {
    startTime = millis();
    running = true;
  }

  void stop() {
    running = false;
  }

  float update() {
    if (isFinished()) {
      running = false;
      value = end;
      return end;
    }

    if (!running) {
      start();
    }

    //scale value
    value = (end - start) * normalizedPhase() + start;

    if (valueArray != null &&
      valueArray.length != 0) {
      for (int i=0; i<valueArray.length; i++) {
        valueArray[i] = (endArray[i] - startArray[i]) * normalizedPhase() + startArray[i];
      }
    }

    return value;
  }

  //================================================================
  //MARK: Normalized Value Calculation
  //================================================================

  float normalizedPhase() {
    if (duration == 0) {
      return 1.0;
    }
    float n = (millis() - startTime) / duration;
    switch (mode) {
    case RAMP_MODE_EASE_IN:
      {
        n *= n;
      }
      break;
    case RAMP_MODE_EASE_OUT:
      {
        n = 1.0 - (1.0 - n) * (1.0 - n);
      }
      break;
    case RAMP_MODE_S_CURVE:
      {
        float dn = 1.0 - pow(1.0 - n, sCurvePower);
        n = 0.5 - (cos(dn * PI) * 0.5);
      }
      break;
    case RAMP_MODE_LINEAR:
    default:
      break;
    }
    n = max(0.0, min(n, 1.0));
    return n;
  }

  //================================================================
  //MARK: Readonly Getters
  //================================================================

  boolean isRunning() {
    return running;
  }

  boolean isFinished() {
    return millis() >= startTime + duration;
  }

  int intValue() {
    return int(value);
  }

  float startValue() {
    return start;
  }

  float endValue() {
    return end;
  }

  //================================================================
  //MARK: Class Functions
  //================================================================

  void randomizeMode() {
    int i = int(random(4));
    switch (i) {
    case 0: 
      mode = RampMode.RAMP_MODE_LINEAR; 
      break;
    case 1: 
      mode = RampMode.RAMP_MODE_EASE_IN;
      break;
    case 2: 
      mode = RampMode.RAMP_MODE_EASE_OUT; 
      break;
    case 3: 
      mode = RampMode.RAMP_MODE_S_CURVE;
      break;
    default:
    }
  }
}