//================================================================
//MARK: Class Segment
//================================================================

class Segment {

  //-------------------------------------------------
  //Public Properties
  //-------------------------------------------------

  float len;
  float theta;
  float bodyPosition, totalBodyLength;
  PVector anchor;

  float segmentStartPhase, segmentEndPhase;
  PVector cp0, cp1;

  boolean drawTendrils = true;

  //-------------------------------------------------
  //Private Properties
  //-------------------------------------------------

  private float _startBodyPosition, _endBodyPosition;
  private Tendril[] tendrils;
  private boolean tendrilsRequireUpdate; //we recalc tendril positions until we get a full phase draw
  private color segmentColor; //use setColor
  private float tendrilSinFrequency = random(2.2, 3.3);
  private color originalColor;

  //-------------------------------------------------
  //Lifecycle
  //-------------------------------------------------

  Segment(float _length, PVector _anchor, PVector _theta, int _tendrilCount) {
    len = _length;
    theta = atan2(_theta.y - _anchor.y, _theta.x - _anchor.x); //note: atan2(y, x);
    anchor = _anchor;

    segmentStartPhase = 0.0;
    segmentEndPhase = 0.0;

    float scx0 = random(len * 0.9) + len * 0.1;
    float scy0 = 10;
    float scx1 = 0;
    float scy1 = 0;

    cp0 = new PVector(scx0, scy0);
    cp1 = new PVector(scx1, scy1);

    //tendrils = new ArrayList<Tendril>();
    tendrils = new Tendril[_tendrilCount];
    //int tendrilCount = int(random(kMinimumTendrilCountPerSegment, kMaximumTendrilCountPerSegment));
    float scale = 1.0 - (_tendrilCount / 20.0);
    for (int i=0; i<_tendrilCount; i++) {
      Side side = (int(random(2)) == 0) ? Side.LEFTSIDE : Side.RIGHTSIDE;
      tendrils[i] = new Tendril(random(kMinimumTendrilLength, kMaximumTendrilLength) * scale, random(1), side);
      //tendrils.add(tendril);
    }

    tendrilsRequireUpdate = true;
  }

  //-------------------------------------------------
  //Update & Draw
  //-------------------------------------------------

  /**
   returns TRUE if update is necessary, FALSE if not
   */
  boolean update(float startBodyPosition, float endBodyPosition) {
    _startBodyPosition = startBodyPosition;
    _endBodyPosition = endBodyPosition;

    //we know our anchorPosition and len, so we can calculate how much of this segment to draw
    //we can also fade in knowing the start and end body positions

    if (startBodyPosition > bodyPosition + len ||
      endBodyPosition < bodyPosition) {
      return false;
    }

    segmentEndPhase = min((endBodyPosition - bodyPosition) / len, 1.0);
    segmentStartPhase = min((max(startBodyPosition, bodyPosition) - bodyPosition) / len, 1.0);

    return true;
  }

  /**
   returns TRUE if update is necessary, FALSE if not
   */
  boolean drawSegment(PGraphics pg) {
    if (_startBodyPosition > bodyPosition + len ||
      _endBodyPosition < bodyPosition) {
      return false;
    }

    if (pg != null) {
      pg.pushMatrix();

      //translate to anchorpoint and rotate
      pg.translate(anchor.x, anchor.y);
      pg.rotate(theta);
    } else {
      pushMatrix();

      //translate to anchorpoint and rotate
      translate(anchor.x, anchor.y);
      rotate(theta);
    }

    if (drawTendrils) {
      for (int i=0; i<tendrils.length; i++) {
        Tendril tendril = tendrils[i];
        if (segmentStartPhase < tendril.phase &&
          segmentEndPhase > tendril.phase) {
          float tendrilBodyPosition = bodyPosition + (tendril.phase * len);
          float endScale = max(min((_endBodyPosition - tendrilBodyPosition) / kBodyTendrilEndFadeLength, 1.0), 0.0);
          float startScale = max(min((tendrilBodyPosition - _startBodyPosition) / kBodyTendrilStartFadeLength, 1.0), 0.0);

          float tendrilScale = startScale * endScale;
          float side = (tendril.side == Side.LEFTSIDE) ? 1.0 : -1.0;
          //float side = 1.0;

          //float frequency = 3.0;
          float tendrilAngle = (tendrilBodyPosition / (totalBodyLength) * PI * tendrilSinFrequency) + PI + PI * 0.5;
          //float tendrilAngle = tendril.phase * TWO_PI;

          if (pg != null) {
            pg.pushMatrix();
          } else {
            pushMatrix();
          }
          if (tendrilsRequireUpdate) {
            updateTendrilPositions();
            if (segmentStartPhase == 0.0 &&
              segmentEndPhase == 1.0) {
              tendrilsRequireUpdate = false;
            }
          }
          if (pg != null) {
            pg.translate(tendril.anchor.x, tendril.anchor.y);
            pg.rotate(-theta + (tendrilAngle * side));
            pg.stroke(tendril.col);
            pg.noFill();
            pg.strokeWeight(kTendrilStrokeWeight);
          } else {
            translate(tendril.anchor.x, tendril.anchor.y);
            rotate(-theta + (tendrilAngle * side));
            stroke(tendril.col);
            noFill();
            strokeWeight(kTendrilStrokeWeight);
          }
          //line(0, 0, tendril.len * tendrilScale * side, 0);

          //tendrilScale = 1.0;
          float scaler = tendrilScale * side;
          float[] tbp = {0, 
            0, 
            tendril.cp0.x * scaler, 
            tendril.cp0.y, 
            tendril.cp1.x * scaler, 
            tendril.cp1.y, 
            tendril.len * scaler, 
            0}; 
          tbp = splitBezier(tbp, tendrilScale);
          if (pg != null) {
            bezierWithPoints(tbp, pg);
          } else {
            bezierWithPoints(tbp);
          }

          if (pg != null) {
            pg.popMatrix();
          } else {
            popMatrix();
          }
        }
      }
    }

    //from here out we consider the body of the segment to be horizontal,
    //tendrils will branch out from top (LEFTSIDE) and bottom (RIGHTSIDE)
    if (pg != null) {
      pg.stroke(segmentColor);
      pg.noFill();
      pg.strokeWeight(kSegmentStrokeWeight);
    } else {
      stroke(segmentColor);
      noFill();
      strokeWeight(kSegmentStrokeWeight);
    }

    float x0 = len * segmentStartPhase;
    float x1 = len * segmentEndPhase;
    float[] bp = {x0, 0, cp0.x, cp0.y, cp1.x, cp1.y, x1, 0};
    //if we're only drawing a portion of the segment...
    if (segmentStartPhase != 0.0 ||
      segmentEndPhase != 1.0) {
      bp = splitBezier(bp, segmentStartPhase, segmentEndPhase);
    }
    if (pg != null) {
      bezierWithPoints(bp, pg);
    } else {
      bezierWithPoints(bp);
    }

    if (pg != null) {
      pg.popMatrix();
    } else {
      popMatrix();
    }
    return true;
  }

  //-------------------------------------------------
  //Functions
  //-------------------------------------------------

  void updateTendrilPositions() {
    float x0 = len * segmentStartPhase;
    float x1 = len * segmentEndPhase;   
    float[] bp = {x0, 0, cp0.x, cp0.y, cp1.x, cp1.y, x1, 0};
    for (int i=0; i<tendrils.length; i++) {
      Tendril tendril = tendrils[i];
      float x = bezierPoint(bp[0], bp[2], bp[4], bp[6], tendril.phase);
      float y = bezierPoint(bp[1], bp[3], bp[5], bp[7], tendril.phase);
      tendril.anchor = new PVector(x, y);
    }
  }

  PVector startPoint() {
    return pointInSegment(segmentStartPhase);
  }

  PVector endPoint() {
    return pointInSegment(segmentEndPhase);
  }

  PVector pointInSegment(float phase) {
    float dx = cos(theta) * (len * phase);
    float dy = sin(theta) * (len * phase);
    return new PVector(anchor.x + dx, anchor.y + dy);
  }

  void setColor(color c) {
    segmentColor = c;
    originalColor = c;
    for (int i=0; i<tendrils.length; i++) {
      tendrils[i].setColor(segmentColor);
    }
  }

  void setColorSaturation(float s) {
    for (int i=0; i<tendrils.length; i++) {
      tendrils[i].setColorSaturation(s);
    }
  }

  void scaleColorBrightness(float b) {
    segmentColor = color(hue(originalColor), saturation(originalColor), brightness(originalColor) * b);
  }

  void setColorBrightness(float b) {
    segmentColor = color(hue(segmentColor), saturation(segmentColor), b);
    originalColor = segmentColor;
  }
}