//Ink Well
//Kevin Schlei, 2017

//================================================================
//MARK: Properties
//================================================================

PVector bodyStartAnchorPoint, bodyEndAnchorPoint;
float kCameraEasedY, kCameraEasedScale;
float bgFade = 1.0;

PImage textureImage;
PImage vignetteImage;

PFont titleFont;

Body[] bodies;

//================================================================
//MARK: Setup & Draw
//================================================================

void setup() {
  //if you want to run in windowed mode, be sure to use P2D
  //size(1280, 960, P2D);

  //fullscreen:
  fullScreen(P2D);
  pixelDensity(displayDensity());
  noCursor();
  frameRate(60);
  colorMode(HSB);

  //preferences
  loadPreferences();

  //vignette corners color setup
  kCornerTintLight = color(20, 255, 80); //rust red
  kCornerTintDark = color(190, 115, 80); //purple glow

  //title setup
  titleFont = createFont("Economica-Regular-OTF.otf", 40);
  textFont(titleFont);

  //background texture images
  vignetteImage = loadImage("clouds-no-vignette.png");
  textureImage = loadImage("ink-bg-texture-sand-white.png");

  //create creature bodies
  bodies = new Body[kBodyCount];
  for (int i=0; i<bodies.length; i++) {
    color bc = color(red(kBodyColorsHSB[i % kBodyColorsHSB.length]), 
      green(kBodyColorsHSB[i % kBodyColorsHSB.length]), 
      blue(kBodyColorsHSB[i % kBodyColorsHSB.length])
      );
    Body b = new Body(kSegmentCount, bc);
    bodies[i] = b;
  }

  //starting camera zoom
  kCameraEasedScale = 0.25;
}

void draw() {
  drawBackground();
  updateBodies();
  updateCamera();
  drawBodies();
}

//-------------------------------------------------
//drawBackground()
//fades solid color background and draws texture images
//-------------------------------------------------
void drawBackground() {
  //background solid color
  if (millis() - kActiveStartTime > kPieceTimeInSeconds * 1000 &&
    kActive) {
    //main piece timing over, fading background and segments to black
    color c = color(hue(kBGColor), saturation(kBGColor), brightness(kBGColor) * bgFade);
    bgFade -= (1.0 / frameRate) / 20.0;
    bgFade = max(0, bgFade);
    background(c);

    for (int i=0; i<bodies.length; i++) {
      bodies[i].scaleSegmentColorBrightness(bgFade);
    }
  } else {
    background(kBGColor);
  }

  //vignette
  beginShape();
  noStroke();
  texture(vignetteImage);
  if (bgFade < 1.0) {
    colorMode(RGB);
    float r = red(kCornerTintLight) * bgFade + red(kCornerTintDark) * (1.0 - bgFade);
    float g = green(kCornerTintLight) * bgFade + green(kCornerTintDark) * (1.0 - bgFade);
    float b = blue(kCornerTintLight) * bgFade + blue(kCornerTintDark) * (1.0 - bgFade);
    tint(r, g, b);
    colorMode(HSB);
  } else {
    tint(kCornerTintLight);
  }
  textureWrap(REPEAT);
  textureMode(NORMAL);
  vertex(0, 0, 0, 0);
  vertex(width, 0, 1, 0);
  vertex(width, height, 1, 1);
  vertex(0, height, 0, 1);
  endShape();
  noTint();
}

//-------------------------------------------------
//updateBodies()
//calls update() on each body, also updates camera zoom
//-------------------------------------------------
void updateBodies() {
  //update bodies
  for (int i=0; i<bodies.length; i++) {
    Body b = bodies[i];
    b.update();
  }
}

void updateCamera() {
  //update camera
  //update / draw bodies
  float maxY = -1;
  float minY = MAX_FLOAT;
  for (int i=0; i<bodies.length; i++) {
    Body b = bodies[i];
    if (b.bodyStartPoint != null &&
      b.bodyStartPoint.y < minY) {
      minY = b.bodyStartPoint.y;
    }
    if (b.bodyEndPoint != null &&
      b.bodyEndPoint.y > maxY) {
      maxY = b.bodyEndPoint.y;
    }
  }

  //the smaller kCameraEasedScale, the more zoomed out
  float heightScale = abs(height / (maxY - minY));
  float widthScale = abs(width / (kBodySpacing * kBodyCount)) * 1.6;
  float targetScale = min(heightScale, widthScale);
  targetScale = min(targetScale, kMaximumCameraZoom);
  float dScale = targetScale - kCameraEasedScale;
  float ease = (targetScale < kCameraEasedScale) ? 0.03 : 0.01;
  kCameraEasedScale = kCameraEasedScale + (dScale * ease);

  //camera movement
  float targetY = (maxY + minY) / 2.0;
  float dy = targetY - kCameraEasedY;
  if (frameCount < 20) {
    kCameraEasedY = kCameraEasedY + (dy);
  } else {
    kCameraEasedY = kCameraEasedY + (dy * 0.01);
  }
}

//-------------------------------------------------
//drawBodies()
//zooms camera, draws bodies and text
//-------------------------------------------------
void drawBodies() {
  pushMatrix();

  if (kRotate90) {
    translate(width / 2 + -kCameraEasedY * kCameraEasedScale, height / 2);
    rotate(PI / 2.0);
  } else {
    translate(width / 2, height / 2 + kCameraEasedY * kCameraEasedScale);
  }
  scale(kCameraEasedScale, -kCameraEasedScale);

  beginShape();
  noStroke();
  texture(textureImage);
  tint(20, 255, 80, 15);
  textureWrap(REPEAT);
  textureMode(NORMAL);
  pushMatrix();
  rotate(PI / 2.0);
  float sc = 2.5;
  vertex(-width * sc, -height * sc, 0, 0);
  vertex(-width * sc, height * sc, 1, 0);
  vertex(width * sc, height * sc, 1, 1);
  vertex(width * sc, -height * sc, 0, 1);
  endShape();
  popMatrix();
  noTint();

  //draw bodies
  if (!kShowInstructions || kActive) {
    for (int i=0; i<bodies.length; i++) {
      Body b = bodies[i];
      float offset = (kBodyCount % 2 == 0) ? kBodySpacing / 2.0 + ((kBodyCount - 1) / 2) * kBodySpacing : ((kBodyCount - 1) / 2) * kBodySpacing;
      pushMatrix();
      translate(i * kBodySpacing - offset, 0);
      b.drawBody(null);
      popMatrix();
    }
  }

  //text
  if (!kActive) {
    textSize(40);
    textAlign(CENTER);
    noStroke();
    color c = color(hue(kBGColor), saturation(kBGColor), brightness(kBGColor) * 0.75);
    fill(c);
    pushMatrix();
    rotate(PI);
    translate(0, height / 2.0);
    scale(-1, 1);
    text("press spacebar to begin\npress 'i' for instructions", 0, 0);
    //text(kInstructionsString, 0, 0);
    popMatrix();
  }
  if (!kActive && kShowInstructions) {
    textSize(40);
    textAlign(CENTER);
    noStroke();
    color c = color(hue(kBGColor), saturation(kBGColor), brightness(kBGColor) * 0.65);
    fill(c);
    pushMatrix();
    rotate(PI);
    scale(-1, 1);
    text(kInstructionsString, 0, 0);
    popMatrix();
  }

  popMatrix();
}

//================================================================
//MARK: Functions
//================================================================

void loadPreferences() {
  String preferences[] = loadStrings("ink_well_preferences.txt");
  for (int i=0; i<preferences.length; i++) {
    String keyValue[] = split(preferences[i], '=');

    //kBodyCount
    if (keyValue[0].equals("creatureCount")) {
      kBodyCount = Integer.parseInt(keyValue[1]);
    }
    
    //kBodySpacing
    if (keyValue[0].equals("creatureSpacing")) {
      kBodySpacing = Float.parseFloat(keyValue[1]);
    }

    //kPieceTimeInSeconds
    if (keyValue[0].equals("duration")) {
      kPieceTimeInSeconds = Integer.parseInt(keyValue[1]);
    }

    //kRotate90
    if (keyValue[0].equals("screenRotate")) {
      kRotate90 = Boolean.parseBoolean(keyValue[1]);
    }
  }
}